# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

import logging
import fnmatch
import os
import re

import gv
import yaml

from kapidox.depdiagram import gvutils
from kapidox.depdiagram.framework import Framework


TARGET_SHAPES = [
    "polygon", # lib
    "house",   # executable
    "octagon", # module (aka plugin)
    "diamond", # static lib
    ]

DEPS_SHAPE = "ellipse"

DEPS_BLACKLIST = [
    "-l*", "-W*", # link flags
    "/*", # absolute dirs
    "m", "pthread", "util", "nsl", "resolv", # generic libs
    "*example*", "*demo*", "*test*", "*Test*", "*debug*" # helper targets
    ]


def preprocess(fname):
    graph_handle = gv.read(fname)
    txt = open(fname).read()
    targets = []

    # Replace the generated node names with their label. CMake generates a graph
    # like this:
    #
    # "node0" [ label="KF5DNSSD" shape="polygon"];
    # "node1" [ label="Qt5::Network" shape="ellipse"];
    # "node0" -> "node1"
    #
    # And we turn it into this:
    #
    # "KF5DNSSD" [ label="KF5DNSSD" shape="polygon"];
    # "Qt5::Network" [ label="Qt5::Network" shape="ellipse"];
    # "KF5DNSSD" -> "Qt5::Network"
    #
    # Using real framework names as labels makes it possible to merge multiple
    # .dot files.
    for node_handle in gvutils.get_node_list(graph_handle):
        node = gvutils.Node(node_handle)
        label = node.label.replace("KF5::", "")
        if node.shape in TARGET_SHAPES:
            targets.append(label)
        txt = txt.replace('"' + node.name + '"', '"' + label + '"')

    # Sometimes cmake will generate an entry for the target alias, something
    # like this:
    #
    # "node9" [ label="KParts" shape="polygon"];
    # ...
    # "node15" [ label="KF5::KParts" shape="ellipse"];
    # ...
    #
    # After our node renaming, this ends up with a second "KParts" node
    # definition, which we need to get rid of.
    for target in targets:
        rx = r' *"' + target + '".*label="KF5::' + target + '".*shape="ellipse".*;'
        txt = re.sub(rx, '', txt)
    return txt


def _add_extra_dependencies(fw, dct):
    lst = dct.get("framework-dependencies")
    if lst is None:
        return
    for dep in lst:
        fw.add_extra_framework(dep)


class DotFileParser(object):
    def __init__(self, with_qt):
        self._with_qt = with_qt

    def parse(self, tier, dot_file):
        name = os.path.basename(dot_file).replace(".dot", "")
        fw = Framework(tier, name)

        # Preprocess dot files so that they can be merged together.
        dot_data =  preprocess(dot_file)
        self._init_fw_from_dot_data(fw, dot_data, self._with_qt)

        return fw

    def _init_fw_from_dot_data(self, fw, dot_data, with_qt):
        def target_from_node(node):
            return node.name.replace("KF5", "")

        src_handle = gv.readstring(dot_data)

        targets = set()
        for node_handle in gvutils.get_node_list(src_handle):
            node = gvutils.Node(node_handle)
            if node.shape in TARGET_SHAPES and self._want(node):
                target = target_from_node(node)
                targets.add(target)
                fw.add_target(target)

        for edge_handle in gvutils.get_edge_list(src_handle):
            edge = gvutils.Edge(edge_handle)
            target = target_from_node(edge.tail)
            if target in targets and self._want(edge.head):
                dep_target = target_from_node(edge.head)
                fw.add_target_dependency(target, dep_target)

    def _want(self, node):
        if node.shape not in TARGET_SHAPES and node.shape != DEPS_SHAPE:
            return False
        name = node.name

        for pattern in DEPS_BLACKLIST:
            if fnmatch.fnmatchcase(node.name, pattern):
                return False
        if not self._with_qt and name.startswith("Qt"):
            return False
        return True


class FrameworkDb(object):
    def __init__(self):
        self._fw_list = []
        self._fw_for_target = {}

    def populate(self, dot_files, with_qt=False):
        """
        Init db from dot files
        """
        parser = DotFileParser(with_qt)
        for dot_file in dot_files:
            yaml_file = dot_file.replace(".dot", ".yaml")
            with open(yaml_file) as f:
                dct = yaml.safe_load(f)

            if 'tier' not in dct:
                # This mean it's not a frameworks
                continue

            tier = dct["tier"]
            fw = parser.parse(tier, dot_file)

            _add_extra_dependencies(fw, dct)
            self._fw_list.append(fw)
        self._update_fw_for_target()

    def _update_fw_for_target(self):
        self._fw_for_target = {}
        for fw in self._fw_list:
            for target in fw.get_targets():
                self._fw_for_target[target] = fw

    def find_by_name(self, name):
        for fw in self._fw_list:
            if fw.name == name:
                return fw
        return None

    def remove_unused_frameworks(self, wanted_fw):
        def add_to_set(fw_set, wanted_fw):
            fw_set.add(wanted_fw)

            for target in wanted_fw.get_all_target_dependencies():
                fw = self._fw_for_target.get(target)
                if fw is not None and not fw in fw_set:
                    add_to_set(fw_set, fw)

            for fw_name in wanted_fw.get_extra_frameworks():
                fw = self.find_by_name(fw_name)
                if not fw:
                    logging.warning("Framework {} has an extra dependency on {}, but there is no such framework".format(wanted_fw, fw_name))
                    continue
                if not fw in fw_set:
                    add_to_set(fw_set, fw)

        fw_set = set()
        add_to_set(fw_set, wanted_fw)
        self._fw_list = list(fw_set)

    def find_external_targets(self):
        all_targets = set([])
        fw_targets = set([])
        for fw in self._fw_list:
            fw_targets.update(fw.get_targets())
            all_targets.update(fw.get_all_target_dependencies())
        return all_targets.difference(fw_targets)

    def get_framework_for_target(self, target):
        return self._fw_for_target[target]

    def __iter__(self):
        return iter(self._fw_list)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

import argparse
import sys

from kapidox import depdiagram

DESCRIPTION = """\
"""


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument("-o", "--output", dest="output", default="-",
                        help="Output to FILE", metavar="FILE")

    parser.add_argument("--qt", dest="qt", action="store_true",
                        help="Show Qt libraries")

    parser.add_argument("--detailed", dest="detailed", action="store_true",
                        help="Show targets within frameworks")

    parser.add_argument("--framework", dest="framework",
                        help="Only show dependencies of framework FRAMEWORK", metavar="FRAMEWORK")

    parser.add_argument("dot_files", nargs="+")

    args = parser.parse_args()

    if args.output == "-":
        out = sys.stdout
    else:
        out = open(args.output, "w")

    if depdiagram.generate(out, args.dot_files, framework=args.framework, with_qt=args.qt, detailed=args.detailed):
        return 0
    else:
        return 1


if __name__ == "__main__":
    sys.exit(main())

# vi: ts=4 sw=4 et



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

"""
A set of classes and functions to make it easier to work with Graphviz graphs
"""

import gv


class Node(object):
    def __init__(self, node_handle):
        self.handle = node_handle

    @property
    def name(self):
        return gv.nameof(self.handle)

    @property
    def label(self):
        return gv.getv(self.handle, "label")

    @property
    def shape(self):
        return gv.getv(self.handle, "shape")


class Edge(object):
    def __init__(self, edge_handle):
        self.handle = edge_handle

    @property
    def head(self):
        handle = gv.headof(self.handle)
        if handle is None:
            return None
        else:
            return Node(handle)

    @property
    def tail(self):
        handle = gv.tailof(self.handle)
        if handle is None:
            return None
        else:
            return Node(handle)


def get_node_list(graph_h):
    """Generator to iterate over all nodes of a graph"""
    handle = gv.firstnode(graph_h)
    while gv.ok(handle):
        yield handle
        handle = gv.nextnode(graph_h, handle)


def get_edge_list(graph_h):
    """Generator to iterate over all edges of a graph"""
    handle = gv.firstedge(graph_h)
    while gv.ok(handle):
        yield handle
        handle = gv.nextedge(graph_h, handle)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

import argparse
import os
import logging
import shutil
import subprocess
import sys
import tempfile
import time

from kapidox import utils

DESCRIPTION = """
Generate Graphviz dot files for one or all frameworks.
"""


def generate_dot(fw_dir, fw_name, output_dir):
    """Calls cmake to generate the dot file for a framework.

    Returns true on success, false on failure"""
    dot_path = os.path.join(output_dir, fw_name + ".dot")
    build_dir = tempfile.mkdtemp(prefix="depdiagram-prepare-build-")
    try:
        ret = subprocess.call(["cmake", fw_dir, "--graphviz={}".format(dot_path)],
                              stdout=open("/dev/null", "w"),
                              cwd=build_dir)
        if ret != 0:
            if os.path.exists(dot_path):
                os.remove(dot_path)
            logging.error("Generating dot file for {} failed.".format(fw_name))
            return False
        # Add a timestamp and version info to help with diagnostics
        with open(dot_path, "a") as f:
            f.write("\n# Generated on {}\n".format(time.ctime()))
            version = utils.get_kapidox_version()
            if version:
                f.write("# By {} {}\n".format(sys.argv[0], version))
    finally:
        shutil.rmtree(build_dir)
    return True


def prepare_one(fw_dir, output_dir):
    fw_name = utils.parse_fancyname(fw_dir)
    if fw_name is None:
        return False

    yaml_path = os.path.join(fw_dir, "metainfo.yaml")
    if not os.path.exists(yaml_path):
        logging.error(f"'{fw_dir}' is not a framework: '{yaml_path}' does not exist.")
        return False

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if not generate_dot(fw_dir, fw_name, output_dir):
        return False
    shutil.copyfile(yaml_path, os.path.join(output_dir, fw_name + ".yaml"))
    return True


def prepare_all(fw_base_dir, dot_dir):
    """Generate dot files for all frameworks.

    Looks for frameworks in `fw_base_dir`. Output the dot files in sub dirs of
    `dot_dir`.
    """
    lst = os.listdir(fw_base_dir)
    fails = []
    for idx, fw_name in enumerate(lst):
        fw_dir = os.path.join(fw_base_dir, fw_name)
        if not os.path.isdir(fw_dir):
            continue

        progress = int(100 * (idx + 1) / len(lst))
        print(f'{progress}% {fw_name}')
        if not prepare_one(fw_dir, dot_dir):
            fails.append(fw_name)
    return fails


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-s", "--single",
                       help="Generate dot files for the framework stored in DIR",
                       metavar="DIR")
    group.add_argument("-a", "--all",
                       help="Generate dot files for all frameworks whose dir is in BASE_DIR",
                       metavar="BASE_DIR")
    parser.add_argument("dot_dir",
                        help="Destination dir where dot files will be generated")

    args = parser.parse_args()

    dot_dir = os.path.abspath(args.dot_dir)

    if args.single:
        fw_dir = os.path.abspath(args.single)
        if prepare_one(fw_dir, dot_dir):
            return 0
        else:
            return 1
    else:
        fw_base_dir = os.path.abspath(args.all)
        fails = prepare_all(fw_base_dir, dot_dir)
        if fails:
            logging.error("{} framework(s) failed: {}".format(len(fails), ", ".join(fails)))
            return 1
        else:
            return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

from functools import reduce

class Framework(object):
    """
    A framework has a tier, a name, a collection of targets it provides.
    For each target it provides, it knows which targets they depend on.

    It also has a list of frameworks it depends on, for framework-to-framework
    dependencies. This is useful for cases when a framework depends on macros
    provided by another framework.
    """
    def __init__(self, tier, name):
        self.tier = tier
        self.name = name

        # A dict of target => set([targets])
        self._target_dict = {}

        # A list of other frameworks (not targets!) this framework depends on.
        self._fw_list = []

    def add_target(self, target):
        # Add a new target provided by this framework
        assert target not in self._target_dict
        self._target_dict[target] = set()

    def add_target_dependency(self, target, dep):
        # Add a new dependency target to a target provided by this framework
        assert target in self._target_dict
        self._target_dict[target].add(dep)

    def get_targets(self):
        return self._target_dict.keys()

    def has_target(self, target):
        return target in self._target_dict

    def get_all_target_dependencies(self):
        deps = self._target_dict.values()
        if not deps:
            return set()
        return reduce(lambda x,y: x.union(y), deps)

    def get_dependencies_for_target(self, target):
        return self._target_dict[target]

    def add_extra_framework(self, name):
        # Add a framework this framework depends on because of a
        # framework-to-framework dependency
        self._fw_list.append(name)

    def get_extra_frameworks(self):
        # Returns frameworks this framework depends on because of a
        # framework-to-framework dependency
        return self._fw_list

    def __repr__(self):
        return self.name



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

import argparse
import logging
import os
import sys


def normalized_path(inputpath):
    return os.path.normpath(inputpath)


def parse_args(depdiagram_available):
    import textwrap
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''Generate API documentation of complex projects.

>> This function must be run from an empty directory (where the documentation will be built).''')
    )
    group = add_sources_group(parser)
    group.add_argument('sourcesdir', type=normalized_path,
                       help='Location of the sources.')
    group.add_argument('--depdiagram-dot-dir', type=normalized_path,
                       help='Generate dependency diagrams, using the .dot files from DIR.',
                       metavar="DIR")
    add_output_group(parser)
    add_qt_doc_group(parser)
    add_paths_group(parser)
    add_misc_group(parser)
    args = parser.parse_args()
    check_common_args(args)

    if args.depdiagram_dot_dir and not depdiagram_available:
        logging.error('You need to install the Graphviz Python bindings to '
                      'generate dependency diagrams.\n'
                      'See <https://www.graphviz.org/download/>.')
        exit(1)

    if not os.path.isdir(args.sourcesdir):
        logging.error(args.sourcesdir + " is not a directory")
        exit(2)

    return args


def add_sources_group(parser):
    group = parser.add_argument_group('sources')
    group.add_argument('--accountsfile', type=normalized_path,
                       help='File with accounts information of SVN contributors.')
    return group


def add_output_group(parser):
    group = parser.add_argument_group('output options')
    group.add_argument('--title', default='API Documentation',
                       help='String to use for page titles.')
    group.add_argument('--man-pages', action='store_true',
                       help='Generate man page documentation.')
    group.add_argument('--qhp', action='store_true',
                       help='Generate Qt Compressed Help documentation.')
    return group


def add_qt_doc_group(parser):
    group = parser.add_argument_group('Qt documentation')
    group.add_argument('--qtdoc-dir', type=normalized_path,
                       help='Location of (local) Qt documentation; this is searched ' +
                            'for tag files to create links to Qt classes.')
    group.add_argument('--qtdoc-link',
                       help='Override Qt documentation location for the links in the ' +
                            'html files.  May be a path or URL.')
    group.add_argument('--qtdoc-flatten-links', action='store_true',
                       help='Whether to assume all Qt documentation html files ' +
                            'are immediately under QTDOC_LINK (useful if you set ' +
                            'QTDOC_LINK to the online Qt documentation).  Ignored ' +
                            'if QTDOC_LINK is not set.')
    return group


def add_paths_group(parser):
    group = parser.add_argument_group('paths')
    group.add_argument('--doxygen', default='doxygen', type=normalized_path,
                       help='(Path to) the doxygen executable.')
    group.add_argument('--qhelpgenerator', default='qhelpgenerator', type=normalized_path,
                       help='(Path to) the qhelpgenerator executable.')
    return group


def add_misc_group(parser):
    scriptdir = os.path.dirname(os.path.realpath(__file__))
    doxdatadir = os.path.join(scriptdir, 'data')

    group = parser.add_argument_group('misc')
    group.add_argument('--doxdatadir', default=doxdatadir, type=normalized_path,
                       help='Location of the HTML header files and support graphics.')
    group.add_argument('--keep-temp-dirs', action='store_true',
                       help='Do not delete temporary dirs, useful for debugging.')
    return parser


def check_common_args(args):
    if not _is_doxdatadir(args.doxdatadir):
        logging.error(f'{args.doxdatadir} is not a valid doxdatadir')
        sys.exit(1)


def _is_doxdatadir(directory):
    for name in ['header.html', 'footer.html', 'htmlresource']:
        if not os.path.exists(os.path.join(directory, name)):
            return False
    return True



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

from fnmatch import fnmatch
import logging
import os
import re
import subprocess
import shutil
import sys
import tempfile
import requests


## @package kapidox.utils
#
# Multiple usage utils.
#
# This module contains code which is shared between depdiagram-prepare and
# other components.
#
# Code in this dir should not import any module which is not shipped with
# Python because this module is used by depdiagram-prepare, which must be able
# to run on builds.kde.org, which may not have all the required dependencies.
#


def setup_logging():
    FORMAT = '%(asctime)s %(levelname)s %(message)s'
    logging.basicConfig(format=FORMAT, datefmt='%H:%M:%S', level=logging.DEBUG)


def tolist(a):
    """ Return a list based on `a`. """
    return a if type(a) is list else [a]


def serialize_name(name):
    """ Return a serialized name.

    For now it only replaces ' ' with '_' and lower the letters.
    """
    if name is not None:
        return '_'.join(name.lower().split(' '))
    else:
        return None


def set_repopath(id):
    """ Return the repopath for the repo id, queried from projects.kde.org

    Args:
        id: unique KDE repo identifier
    """
    if id is None:
        return None

    try:
        r = requests.get('https://projects.kde.org/api/v1/identifier/' + id)
        return r.json()['repo']
    except Exception as exc:
        # Catch all exceptions here: whatever fails in this function should not
        # cause the code to fail
        logging.warning("Failed to get repository url for '{}' from projects.kde.org: {}".format(id, exc))
        # This means there is no canonical repo identifier for this repo:
        # generally that means that the repo was checked out into a non-
        # canonical pathname (e.g. kitemviews checkout out into a directory
        # called KItemViews, or kitemviews.git .. anything other than
        # kitemviews is not recognized).
        return None


def set_maintainers(maintainer_keys, all_maintainers):
    """ Expend the name of the maintainers.

    Args:
        maintainer_keys: (string) Key of the dictionary where the name to expend is saved.
        all_maintainers: (dict of dict) Look-up table where the names and emails of
    the maintainers are stored.

    Examples:
        >>> maintainer_keys = ['arthur', 'toto']

        >>> myteam = {'arthur': {'name': 'Arthur Pendragon',
                                 'email': 'arthur@example.com'},
                      'toto': {'name': 'Toto',
                               'email: 'toto123@example.com'}
                      }

        >>> set_maintainers(maintainer_keys, all_maintainers)
    """

    if not maintainer_keys:
        maintainers = []
    elif isinstance(maintainer_keys, list):
        maintainers = map(lambda x: all_maintainers.get(x, None),
                          maintainer_keys)
    else:
        maintainers = [all_maintainers.get(maintainer_keys, None)]

    maintainers = [x for x in maintainers if x is not None]
    return maintainers


def parse_fancyname(fw_dir):
    """Return the framework name for a given source dir

    The framework name is the name of the toplevel CMake project
    """
    cmakelists_path = os.path.join(fw_dir, "CMakeLists.txt")
    if not os.path.exists(cmakelists_path):
        for f in os.listdir(fw_dir):
            if ".qbs" in f and "Test" not in f:
                return f[:-4]
        logging.error("No CMakeLists.txt in {}".format(fw_dir))
        return None
    project_re = re.compile(r"^\s*project\s*\(\s*([\w\-\_]+)", re.I | re.M)
    with open(cmakelists_path) as f:
        cmakelists_content = f.read()
        match = project_re.search(cmakelists_content)
        if match:
            return match.group(1)

    logging.error(f"Failed to find framework name: Could not find a 'project()' command in {cmakelists_path}.")
    return None


def cache_dir():
    """Find/create a semi-long-term cache directory.

    We do not use tempdir, except as a fallback, because temporary directories
    are intended for files that only last for the program's execution.
    """
    cachedir = None
    if sys.platform == 'darwin':
        try:
            from AppKit import NSSearchPathForDirectoriesInDomains
            # http://developer.apple.com/DOCUMENTATION/Cocoa/Reference/Foundation/Miscellaneous/Foundation_Functions/Reference/reference.html#//apple_ref/c/func/NSSearchPathForDirectoriesInDomains
            # NSApplicationSupportDirectory = 14
            # NSUserDomainMask = 1
            # True for expanding the tilde into a fully qualified path
            cachedir = os.path.join(
                NSSearchPathForDirectoriesInDomains(14, 1, True)[0],
                'KApiDox')
        except:
            pass
    elif os.name == "posix":
        if 'HOME' in os.environ and os.path.exists(os.environ['HOME']):
            cachedir = os.path.join(os.environ['HOME'], '.cache', 'kapidox')
    elif os.name == "nt":
        if 'APPDATA' in os.environ and os.path.exists(os.environ['APPDATA']):
            cachedir = os.path.join(os.environ['APPDATA'], 'KApiDox')
    if cachedir is None:
        cachedir = os.path.join(tempfile.gettempdir(), 'kapidox')
    if not os.path.isdir(cachedir):
        os.makedirs(cachedir)
    return cachedir


def svn_export(remote, local, overwrite=False):
    """Wraps svn export.

    Args:
        remote:     (string) the remote url.
        local:      (string) the local path where to download.
        overwrite:  (bool) whether to overwrite `local` or not. (optional,
    default = False)

    Returns:
        True if success.

    Raises:
        FileNotFoundError: &nbsp;
        subprocess.CalledProcessError: &nbsp;
    """
    try:
        import svn.core
        import svn.client
        logging.debug("Using Python libsvn bindings to fetch %s", remote)
        ctx = svn.client.create_context()
        ctx.auth_baton = svn.core.svn_auth_open([])

        latest = svn.core.svn_opt_revision_t()
        latest.type = svn.core.svn_opt_revision_head

        svn.client.export(remote, local, latest, True, ctx)
    except ImportError:
        logging.debug("Using external svn client to fetch %s", remote)
        cmd = ['svn', 'export', '--quiet']
        if overwrite:
            cmd.append('--force')
        cmd += [remote, local]
        try:
            subprocess.check_call(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise subprocess.StandardException(e.output)
        except FileNotFoundError as e:
            logging.debug("External svn client not found")
            return False
    # subversion will set the timestamp to match the server
    os.utime(local, None)
    return True


def copy_dir_contents(directory, dest):
    """Copy the contents of a directory

    Args:
        directory: (string) the directory to copy the contents of.
        dest: (string) the directory to copy them into.
    """
    ignored = ['CMakeLists.txt']
    ignore = shutil.ignore_patterns(*ignored)
    for fn in os.listdir(directory):
        f = os.path.join(directory, fn)
        if os.path.isfile(f):
            docopy = True
            for i in ignored:
                if fnmatch(fn, i):
                    docopy = False
                    break
            if docopy:
                shutil.copy(f, dest)
        elif os.path.isdir(f):
            dest_f = os.path.join(dest, fn)
            if os.path.isdir(dest_f):
                shutil.rmtree(dest_f)
            shutil.copytree(f, dest_f, ignore=ignore)


_KAPIDOX_VERSION = None


def get_kapidox_version():
    """Get commit id of running code if it is running from git repository.

    May return an empty string if it failed to extract the commit id.

    Assumes .git/HEAD looks like this:

        ref: refs/heads/master

    and assumes .git/refs/heads/master contains the commit id
    """
    global _KAPIDOX_VERSION

    if _KAPIDOX_VERSION is not None:
        return _KAPIDOX_VERSION

    _KAPIDOX_VERSION = ""
    bin_dir = os.path.dirname(sys.argv[0])
    git_dir = os.path.join(bin_dir, "..", ".git")
    if not os.path.isdir(git_dir):
        # Looks like we are not running from the git repo, exit silently
        return _KAPIDOX_VERSION

    git_HEAD = os.path.join(git_dir, "HEAD")
    if not os.path.isfile(git_HEAD):
        logging.warning(f'Getting git info failed: {git_HEAD} is not a file')
        return _KAPIDOX_VERSION

    try:
        line = open(git_HEAD).readline()
        ref_name = line.split(": ")[1].strip()
        with open(os.path.join(git_dir, ref_name)) as f:
            _KAPIDOX_VERSION = f.read().strip()
    except Exception as exc:
        # Catch all exceptions here: whatever fails in this function should not
        # cause the code to fail
        logging.warning(f'Getting git info failed: {exc}')
    return _KAPIDOX_VERSION


def find_dot_files(dot_dir):
    """Returns a list of path to files ending with .dot in subdirs of `dot_dir`."""
    lst = []
    for (root, dirs, files) in os.walk(dot_dir):
        lst.extend([os.path.join(root, x) for x in files if x.endswith('.dot')])
    return lst



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
# SPDX-FileCopyrightText: 2014 Alex Merry <alex.merry@kdemail.net>
# SPDX-FileCopyrightText: 2014 Alex Turbov <i.zaufi@gmail.com>
#
# SPDX-License-Identifier: BSD-2-Clause

def _quote(txt):
    return '"' + txt + '"'


class DoxyfileWriter(object):
    """Makes it easy to write entries in a Doxyfile, correctly handling quoting
    """
    def __init__(self, fl):
        self.fl = fl

    def write_entry(self, key, value):
        """Write an entry

        Args:
            key: the key part of the entry
            value: the value part of the entry. Can be a string, a list, a
        tuple or a boolean
        """
        if isinstance(value, (list, tuple)):
            txt = ' '.join([_quote(x) for x in value])
        elif isinstance(value, bool):
            txt = ['NO', 'YES'][value]
        else:
            txt = _quote(str(value))
        self.fl.write(key + ' = ' + txt + '\n')

    def write_entries(self, **kwargs):
        """ Call write_entry for all arguments passed"""
        for key, value in kwargs.items():
            self.write_entry(key, value)



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Alex Merry <alex.merry@kdemail.net>
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
# SPDX-FileCopyrightText: 2014 Alex Turbov <i.zaufi@gmail.com>
# SPDX-FileCopyrightText: 2016 Olivier Churlaud <olivier@churlaud.com>
#
# SPDX-License-Identifier: BSD-2-Clause

import codecs
from distutils.spawn import find_executable
import datetime
import os
import logging
from os import environ
import shutil
import subprocess
import tempfile
import sys
import pathlib
from typing import Any, Dict
import xml.etree.ElementTree as ET
import re
import glob
from pathlib import Path

import jinja2

from urllib.parse import urljoin

import xml.etree.ElementTree as xmlET
import json

from jinja2.environment import Template

from kapidox import utils
try:
    from kapidox import depdiagram
    DEPDIAGRAM_AVAILABLE = True
except ImportError:
    DEPDIAGRAM_AVAILABLE = False

from .doxyfilewriter import DoxyfileWriter


# @package kapidox.generator
#
# The generator

__all__ = (
    "Context",
    "generate_apidocs",
    "search_for_tagfiles",
    "WARN_LOGFILE",
    "build_classmap",
    "postprocess",
    "create_dirs",
    "create_jinja_environment",
    )

WARN_LOGFILE = 'doxygen-warnings.log'

HTML_SUBDIR = 'html'


class Context(object):
    """
    Holds parameters used by the various functions of the generator
    """
    __slots__ = (
        # Names
        'modulename',
        'fancyname',
        'title',
        'fwinfo',
        # KApidox files
        'doxdatadir',
        'resourcedir',
        # Input
        'srcdir',
        'tagfiles',
        'dependency_diagram',
        'copyright',
        'is_qdoc',
        # Output
        'outputdir',
        'htmldir',
        'tagfile',
        # Output options
        'man_pages',
        'qhp',
        # Binaries
        'doxygen',
        'qhelpgenerator',
    )

    def __init__(self, args, **kwargs):
        # Names
        self.title = args.title
        # KApidox files
        self.doxdatadir = args.doxdatadir
        # Output options
        self.man_pages = args.man_pages
        self.qhp = args.qhp
        # Binaries
        self.doxygen = args.doxygen
        self.qhelpgenerator = args.qhelpgenerator

        for key in self.__slots__:
            if not hasattr(self, key):
                setattr(self, key, kwargs.get(key))


def create_jinja_environment(doxdatadir):
    loader = jinja2.FileSystemLoader(os.path.join(doxdatadir, 'templates'))
    return jinja2.Environment(loader=loader)


def process_toplevel_html_file(outputfile, doxdatadir, products, title, qch_enabled=False):

    def sort_product(product):
        print(product.fancyname)
        if product.fancyname == "The KDE Frameworks":
            return 'aa'
        if product.fancyname == "KDE PIM":
            return 'ab'
        return product.fancyname.lower()

    products.sort(key=sort_product)
    mapping = {
            'resources': './resources',
            # steal the doxygen css from one of the frameworks
            # this means that all the doxygen-provided images etc. will be found
            'title': title,
            'qch': qch_enabled,
            'breadcrumbs': {
                'entries': [
                    {
                        'href': './index.html',
                        'text': 'KDE API Reference'
                    }
                    ]
                },
            'product_list': products,
        }
    tmpl = create_jinja_environment(doxdatadir).get_template('frontpage.html')
    with codecs.open(outputfile, 'w', 'utf-8') as outf:
        outf.write(tmpl.render(mapping))

    tmpl2 = create_jinja_environment(doxdatadir).get_template('search.html')
    search_output = "search.html"
    with codecs.open(search_output, 'w', 'utf-8') as outf:
        outf.write(tmpl2.render(mapping))


def process_subgroup_html_files(outputfile, doxdatadir, groups, available_platforms, title, qch_enabled=False):

    for group in groups:
        mapping = {
            'resources': '../resources',
            'title': title,
            'qch': qch_enabled,
            'breadcrumbs': {
                'entries': [
                    {
                        'href': '../index.html',
                        'text': 'KDE API Reference'
                    },
                    {
                        'href': './index.html',
                        'text': group.fancyname
                    }
                    ]
                },
            'group': group,
            'available_platforms': sorted(available_platforms),
        }

        if not os.path.isdir(group.name):
            os.mkdir(group.name)
        outputfile = group.name + '/index.html'
        tmpl = create_jinja_environment(doxdatadir).get_template('subgroup.html')
        with codecs.open(outputfile, 'w', 'utf-8') as outf:
            outf.write(tmpl.render(mapping))

        tmpl2 = create_jinja_environment(doxdatadir).get_template('search.html')
        search_output = group.name + "/search.html"
        with codecs.open(search_output, 'w', 'utf-8') as outf:
            outf.write(tmpl2.render(mapping))


def create_dirs(ctx):
    ctx.htmldir = os.path.join(ctx.outputdir, HTML_SUBDIR)
    ctx.tagfile = os.path.join(ctx.htmldir, ctx.fwinfo.fancyname + '.tags')

    if not os.path.exists(ctx.outputdir):
        os.makedirs(ctx.outputdir)

    if os.path.exists(ctx.htmldir):
        # If we have files left there from a previous run but which are no
        # longer generated (for example because a C++ class has been removed)
        # then postprocess will fail because the left-over file has already been
        # processed. To avoid that, we delete the html dir.
        shutil.rmtree(ctx.htmldir)
    os.makedirs(ctx.htmldir)


def load_template(path):
    # Set errors to 'ignore' because we don't want weird characters in Doxygen
    # output (for example source code listing) to cause a failure
    content = codecs.open(path, encoding='utf-8', errors='ignore').read()
    try:
        return jinja2.Template(content)
    except jinja2.exceptions.TemplateSyntaxError:
        logging.error('Failed to parse template {}'.format(path))
        raise


def find_tagfiles(docdir, doclink=None, flattenlinks=False, exclude=None, _depth=0):
    """Find Doxygen-generated tag files in a directory.

    The tag files must have the extension .tags, and must be in the listed
    directory, a subdirectory or a subdirectory named html of a subdirectory.

    Args:
        docdir:       (string) the directory to search.
        doclink:      (string) the path or URL to use when creating the
                      documentation links; if None, this will default to
                      docdir. (optional, default None)
        flattenlinks: (bool) if True, generated links will assume all the html
                      files are directly under doclink; else the html files are
                      assumed to be at the same relative location to doclink as
                      the tag file is to docdir; ignored if doclink is not set.
                      (optional, default False)

    Returns:
        A list of pairs of (tag_file,link_path).
    """

    if not os.path.isdir(docdir):
        return []

    if doclink is None:
        doclink = docdir
        flattenlinks = False

    def smartjoin(pathorurl1, *args):
        """Join paths or URLS

        It figures out which it is from whether the first contains a "://"
        """
        if '://' in pathorurl1:
            if not pathorurl1.endswith('/'):
                pathorurl1 += '/'
            return urljoin(pathorurl1, *args)
        else:
            return os.path.join(pathorurl1, *args)

    def nestedlink(subdir):
        if flattenlinks:
            return doclink
        else:
            return smartjoin(doclink, subdir)

    tagfiles = []

    entries = os.listdir(docdir)
    for e in entries:
        if e == exclude:
            continue
        path = os.path.join(docdir, e)
        if os.path.isfile(path) and e.endswith('.tags'):
            tagfiles.append((path, doclink))
        elif (_depth == 0 or (_depth == 1 and e == 'html')) and os.path.isdir(path):
            tagfiles += find_tagfiles(path, nestedlink(e),
                                      flattenlinks=flattenlinks,
                                      _depth=_depth+1,
                                      exclude=exclude)

    return tagfiles


def search_for_tagfiles(suggestion=None, doclink=None, flattenlinks=False, searchpaths=[], exclude=None):
    """Find Doxygen-generated tag files

    See the find_tagfiles documentation for how the search is carried out in
    each directory; this just allows a list of directories to be searched.

    At least one of docdir or searchpaths must be given for it to find anything.

    Args:
        suggestion:   the first place to look (will complain if there are no
                      documentation tag files there)
        doclink:      the path or URL to use when creating the documentation
                      links; if None, this will default to docdir
        flattenlinks: if this is True, generated links will assume all the html
                      files are directly under doclink; if False (the default),
                      the html files are assumed to be at the same relative
                      location to doclink as the tag file is to docdir; ignored
                      if doclink is not set
        searchpaths:  other places to look for documentation tag files

    Returns:
        A list of pairs of (tag_file,link_path)
    """

    if suggestion is not None:
        if not os.path.isdir(suggestion):
            logging.warning(suggestion + " is not a directory")
        else:
            tagfiles = find_tagfiles(suggestion, doclink, flattenlinks, exclude)
            if len(tagfiles) == 0:
                logging.warning(suggestion + " does not contain any tag files")
            else:
                return tagfiles

    for d in searchpaths:
        tagfiles = find_tagfiles(d, doclink, flattenlinks, exclude)
        if len(tagfiles) > 0:
            logging.info("Documentation tag files found at " + d)
            return tagfiles

    return []


def menu_items(htmldir, modulename):
    """Menu items for standard Doxygen files

    Looks for a set of standard Doxygen files (like namespaces.html) and
    provides menu text for those it finds in htmldir.

    Args:
        htmldir:    (string) the directory the HTML files are contained in.
        modulename: (string) the name of the library

    Returns:
        A list of maps with 'text' and 'href' keys.
    """
    entries = [
            {'text': 'Main Page', 'href': 'index.html'},
            {'text': 'Namespace List', 'href': 'namespaces.html'},
            {'text': 'Namespace Members', 'href': 'namespacemembers.html'},
            {'text': 'Alphabetical List', 'href': 'classes.html'},
            {'text': 'Class List', 'href': 'annotated.html'},
            {'text': 'Class Hierarchy', 'href': 'hierarchy.html'},
            {'text': 'File List', 'href': 'files.html'},
            {'text': 'File Members', 'href': 'globals.html'},
            {'text': 'Modules', 'href': 'modules.html'},
            {'text': 'Directories', 'href': 'dirs.html'},
            {'text': 'Dependencies', 'href': modulename + '-dependencies.html'},
            {'text': 'Related Pages', 'href': 'pages.html'},
            ]
    # NOTE In Python 3 filter() builtin returns an iterable, but not a list
    #      type, so explicit conversion is here!
    return list(filter(
            lambda e: os.path.isfile(os.path.join(htmldir, e['href'])),
            entries))


def parse_dox_html(stream):
    """Parse the HTML files produced by Doxygen, extract the key/value block we
    add through header.html and return a dict ready for the Jinja template.

    The HTML files produced by Doxygen with our custom header and footer files
    look like this:

    @code
    <!--
    key1: value1
    key2: value2
    ...
    -->
    <html>
    <head>
    ...
    </head>
    <body>
    ...
    </body>
    </html>
    @endcode

    The parser fills the dict from the top key/value block, and add the content
    of the body to the dict using the "content" key.

    We do not use an XML parser because the HTML file might not be well-formed,
    for example if the documentation contains raw HTML.

    The key/value block is kept in a comment so that it does not appear in Qt
    Compressed Help output, which is not post processed by ourself.
    """
    dct = {}
    body = []

    def parse_key_value_block(line):
        if line == "<!--":
            return parse_key_value_block
        if line == "-->":
            return skip_head
        key, value = line.split(': ', 1)
        dct[key] = value
        return parse_key_value_block

    def skip_head(line):
        if line == "<body>":
            return extract_body
        else:
            return skip_head

    def extract_body(line):
        if line == "</body>":
            return None
        body.append(line)
        return extract_body

    parser = parse_key_value_block
    while parser is not None:
        line = stream.readline().rstrip()
        parser = parser(line)

    dct['content'] = '\n'.join(body)
    return dct


def postprocess_internal_qdoc(htmldir: str, tmpl: Template, env: Dict[str, Any]):
    """Substitute text in HTML files

    Performs text substitutions on each line in each .html file in a directory.

    Args:
        htmldir: (string) the directory containing the .html files.
        mapping: (dict) a dict of mappings.

    """
    for path in glob.glob(os.path.join(htmldir, "*.html")):
        newpath = f"{path}.new"

        txt = Path(path).read_text()
        env['docs'] = txt.partition('body')[2].partition('</body>')[0]

        with codecs.open(newpath, 'w', 'utf-8') as outf:
            try:
                html = tmpl.render(env)
            except BaseException:
                logging.error(f"Postprocessing {path} failed")
                raise

            outf.write(html)

        os.remove(path)
        os.rename(newpath, path)


def postprocess_internal(htmldir, tmpl, mapping):
    """Substitute text in HTML files

    Performs text substitutions on each line in each .html file in a directory.

    Args:
        htmldir: (string) the directory containing the .html files.
        mapping: (dict) a dict of mappings.

    """
    for name in os.listdir(htmldir):
        if name.endswith('.html'):
            path = os.path.join(htmldir, name)
            newpath = path + '.new'

            if name != 'classes.html' and name.startswith('class'):
                mapping['classname'] = name[5:-5].split('_1_1')[-1]
                mapping['fullname'] = name[5:-5].replace('_1_1', '::')
            elif name.startswith('namespace') and name != 'namespaces.html' and not name.startswith('namespacemembers'):
                mapping['classname'] = None
                mapping['fullname'] = name[9:-5].replace('_1_1', '::')
            else:
                mapping['classname'] = None
                mapping['fullname'] = None

            with codecs.open(path, 'r', 'utf-8', errors='ignore') as f:
                mapping['dox'] = parse_dox_html(f)

            with codecs.open(newpath, 'w', 'utf-8') as outf:
                try:
                    html = tmpl.render(mapping)
                except Exception:
                    logging.error('postprocessing {} failed'.format(path))
                    raise
                outf.write(html)
            os.remove(path)
            os.rename(newpath, path)


def build_classmap(tagfile):
    """Parses a tagfile to get a map from classes to files

    Args:
        tagfile: the Doxygen-generated tagfile to parse.

    Returns:
        A list of maps (keys: classname and filename).
    """
    import xml.etree.ElementTree as ET
    tree = ET.parse(tagfile)
    tagfile_root = tree.getroot()
    mapping = []
    for compound in tagfile_root:
        kind = compound.get('kind')
        if kind == 'class' or kind == 'namespace':
            name_el = compound.find('name')
            filename_el = compound.find('filename')
            mapping.append({'classname': name_el.text,
                            'filename': filename_el.text})
    return mapping


def generate_dependencies_page(tmp_dir, doxdatadir, modulename, dependency_diagram):
    """Create `modulename`-dependencies.md in `tmp_dir`"""
    template_path = os.path.join(doxdatadir, 'dependencies.md.tmpl')
    out_path = os.path.join(tmp_dir, modulename + '-dependencies.md')
    tmpl = load_template(template_path)
    with codecs.open(out_path, 'w', 'utf-8') as outf:
        txt = tmpl.render({
                'modulename': modulename,
                'diagramname': os.path.basename(dependency_diagram),
                })
        outf.write(txt)
    return out_path


def generate_apidocs_qdoc(ctx: Context, tmp_dir: str, doxyfile_entries=None, keep_temp_dirs=False):
    absolute = pathlib.Path(os.path.join(ctx.outputdir, 'html')).absolute()

    environ['KAPIDOX_DIR'] = ctx.doxdatadir

    logging.info(f'Running QDoc (qdoc {ctx.fwinfo.path}/.qdocconf --outputdir={absolute}')
    ret = subprocess.call(['qdoc', ctx.fwinfo.path + "/.qdocconf", f"--outputdir={absolute}"])
    if ret != 0:
        raise Exception("QDoc exited with a non-zero status code")


def generate_apidocs(ctx: Context, tmp_dir, doxyfile_entries=None, keep_temp_dirs=False):
    """Generate the API documentation for a single directory"""

    if ctx.is_qdoc:
        return generate_apidocs_qdoc(ctx, tmp_dir, doxyfile_entries, keep_temp_dirs)

    def find_src_subdir(dirlist, deeper_subd=None):
        returnlist = []
        for d in dirlist:
            pth = os.path.join(ctx.fwinfo.path, d)
            if deeper_subd is not None:
                pth = os.path.join(pth, deeper_subd)
            if os.path.isdir(pth) or os.path.isfile(pth):
                returnlist.append(pth)
            else:
                pass  # We drop it
        return returnlist

    input_list = []
    if os.path.isfile(ctx.fwinfo.path + "/Mainpage.dox"):
        input_list.append(ctx.fwinfo.path + "/Mainpage.dox")
    elif os.path.isfile(ctx.fwinfo.path + "/README.md"):
        input_list.append(ctx.fwinfo.path + "/README.md")

    input_list.extend(find_src_subdir(ctx.fwinfo.srcdirs))
    input_list.extend(find_src_subdir(ctx.fwinfo.docdir))
    image_path_list = []

    if ctx.dependency_diagram:
        input_list.append(generate_dependencies_page(tmp_dir,
                                                     ctx.doxdatadir,
                                                     ctx.modulename,
                                                     ctx.dependency_diagram))
        image_path_list.append(ctx.dependency_diagram)

    doxyfile_path = os.path.join(tmp_dir, 'Doxyfile')
    with codecs.open(doxyfile_path, 'w', 'utf-8') as doxyfile:
        # Global defaults
        with codecs.open(os.path.join(ctx.doxdatadir, 'Doxyfile.global'),
                         'r', 'utf-8') as f:
            for line in f:
                doxyfile.write(line)

        writer = DoxyfileWriter(doxyfile)
        writer.write_entry('PROJECT_NAME', ctx.fancyname)
        # FIXME: can we get the project version from CMake? No from GIT TAGS!

        # Input locations
        image_path_list.extend(find_src_subdir(ctx.fwinfo.docdir, 'pics'))
        writer.write_entries(
                INPUT=input_list,
                DOTFILE_DIRS=find_src_subdir(ctx.fwinfo.docdir, 'dot'),
                EXAMPLE_PATH=find_src_subdir(ctx.fwinfo.exampledirs),
                IMAGE_PATH=image_path_list)

        # Other input settings
        writer.write_entry('TAGFILES', [f + '=' + loc for f, loc in ctx.tagfiles])

        # Output locations
        writer.write_entries(
                OUTPUT_DIRECTORY=ctx.outputdir,
                GENERATE_TAGFILE=ctx.tagfile,
                HTML_OUTPUT=HTML_SUBDIR,
                WARN_LOGFILE=os.path.join(ctx.outputdir, WARN_LOGFILE))

        # Other output settings
        writer.write_entries(
                HTML_HEADER=ctx.doxdatadir + '/header.html',
                HTML_FOOTER=ctx.doxdatadir + '/footer.html'
                )

        # Set a layout so that properties are first
        writer.write_entries(
            LAYOUT_FILE=ctx.doxdatadir + '/DoxygenLayout.xml'
            )

        # Always write these, even if QHP is disabled, in case Doxygen.local
        # overrides it
        writer.write_entries(
                QHP_VIRTUAL_FOLDER=ctx.modulename,
                QHP_NAMESPACE="org.kde." + ctx.modulename,
                QHG_LOCATION=ctx.qhelpgenerator)

        writer.write_entries(
                GENERATE_MAN=ctx.man_pages,
                GENERATE_QHP=ctx.qhp)

        if doxyfile_entries:
            writer.write_entries(**doxyfile_entries)

        # Module-specific overrides
        if find_src_subdir(ctx.fwinfo.docdir):
            localdoxyfile = os.path.join(find_src_subdir(ctx.fwinfo.docdir)[0], 'Doxyfile.local')
            if os.path.isfile(localdoxyfile):
                with codecs.open(localdoxyfile, 'r', 'utf-8') as f:
                    for line in f:
                        doxyfile.write(line)

    logging.info('Running Doxygen')
    subprocess.call([ctx.doxygen, doxyfile_path])


def generate_diagram(png_path, fancyname, dot_files, tmp_dir):
    """Generate a dependency diagram for a framework.
    """
    def run_cmd(cmd, **kwargs):
        try:
            subprocess.check_call(cmd, **kwargs)
        except subprocess.CalledProcessError as exc:
            logging.error(f'Command {exc.cmd} failed with error code {exc.returncode}.')
            return False
        return True

    logging.info('Generating dependency diagram')
    dot_path = os.path.join(tmp_dir, fancyname + '.dot')

    with open(dot_path, 'w') as f:
        with_qt = False
        ok = depdiagram.generate(f, dot_files, framework=fancyname,
                                 with_qt=with_qt)
        if not ok:
            logging.error('Generating diagram failed')
            return False

    logging.info('- Simplifying diagram')
    simplified_dot_path = os.path.join(tmp_dir, fancyname + '-simplified.dot')
    with open(simplified_dot_path, 'w') as f:
        if not run_cmd(['tred', dot_path], stdout=f):
            return False

    logging.info('- Generating diagram png')
    if not run_cmd(['dot', '-Tpng', '-o' + png_path, simplified_dot_path]):
        return False

    # These os.unlink() calls are not in a 'finally' block on purpose.
    # Keeping the dot files around makes it possible to inspect their content
    # when running with the --keep-temp-dirs option. If generation fails and
    # --keep-temp-dirs is not set, the files will be removed when the program
    # ends because they were created in `tmp_dir`.
    os.unlink(dot_path)
    os.unlink(simplified_dot_path)
    return True


def create_fw_context(args, lib, tagfiles, copyright=''):

    # There is one more level for groups
    if lib.part_of_group:
        corrected_tagfiles = []
        for k in range(len(tagfiles)):
            # tagfiles are tuples like:
            # ('/usr/share/doc/qt/KF5Completion.tags', '/usr/share/doc/qt')
            # ('/where/the/tagfile/is/Name.tags', '/where/the/root/folder/is')
            if tagfiles[k][1].startswith("http://") or tagfiles[k][1].startswith("https://"):
                corrected_tagfiles.append(tagfiles[k])
            else:
                corrected_tagfiles.append((tagfiles[k][0], '../' + tagfiles[k][1]))
    else:
        corrected_tagfiles = tagfiles

    return Context(args,
                   # Names
                   modulename=lib.name,
                   fancyname=lib.fancyname,
                   fwinfo=lib,
                   # KApidox files
                   resourcedir=('../../../resources' if lib.part_of_group
                                else '../../resources'),
                   # Input
                   copyright=copyright,
                   tagfiles=corrected_tagfiles,
                   dependency_diagram=lib.dependency_diagram,
                   # Output
                   outputdir=lib.outputdir,
                   is_qdoc=lib.metainfo['qdoc'],
                   )


def gen_fw_apidocs(ctx, tmp_base_dir):
    create_dirs(ctx)
    # tmp_dir is deleted when tmp_base_dir is
    tmp_dir = tempfile.mkdtemp(prefix=ctx.modulename + '-', dir=tmp_base_dir)
    generate_apidocs(ctx, tmp_dir,
                     doxyfile_entries=dict(WARN_IF_UNDOCUMENTED=True)
                     )


def create_fw_tagfile_tuple(lib):
    tagfile = os.path.abspath(
                os.path.join(
                    lib.outputdir,
                    'html',
                    lib.fancyname+'.tags'))
    if lib.part_of_group:
        prefix = '../../'
    else:
        prefix = '../../'
    return tagfile, prefix + lib.outputdir + '/html/'


def finish_fw_apidocs_doxygen(ctx: Context, env: Dict[str, Any]):
    tmpl = create_jinja_environment(ctx.doxdatadir).get_template('library.html')
    postprocess_internal(ctx.htmldir, tmpl, env)

    tmpl2 = create_jinja_environment(ctx.doxdatadir).get_template('search.html')
    search_output = ctx.fwinfo.outputdir + "/html/search.html"
    with codecs.open(search_output, 'w', 'utf-8') as outf:
        outf.write(tmpl2.render(env))


def finish_fw_apidocs_qdoc(ctx: Context, env: Dict[str, Any]):
    tmpl = create_jinja_environment(ctx.doxdatadir).get_template('qdoc-wrapper.html')
    postprocess_internal_qdoc(ctx.htmldir, tmpl, env)


def gen_template_environment(ctx: Context) -> Dict[str, Any]:
    classmap = build_classmap(ctx.tagfile)

    entries = [{
        'href': '../../index.html',
        'text': 'KDE API Reference'
    }]

    if ctx.fwinfo.part_of_group:
        entries[0]['href'] = '../' + entries[0]['href']
        entries.append({'href': '../../index.html', 'text': ctx.fwinfo.product.fancyname })

    entries.append({'href': 'index.html', 'text': ctx.fancyname })

    mapping = {
        'qch': ctx.qhp,
        'doxygencss': 'doxygen.css',
        'resources': ctx.resourcedir,
        'title': ctx.title,
        'fwinfo': ctx.fwinfo,
        'copyright': f"1996-{datetime.date.today().year} The KDE developers",
        'doxygen_menu': {'entries': menu_items(ctx.htmldir, ctx.modulename)},
        'class_map': {'classes': classmap},
        'kapidox_version': utils.get_kapidox_version(),
        'breadcrumbs': {
            'entries': entries
        },
    }

    return mapping


def finish_fw_apidocs(ctx: Context):
    env = gen_template_environment(ctx)

    if ctx.is_qdoc:
        logging.info('Postprocessing QtDoc...')

        finish_fw_apidocs_qdoc(ctx, env)

    else:
        logging.info('Postprocessing Doxygen...')

        finish_fw_apidocs_doxygen(ctx, env)


def indexer(lib):
    """ Create json index from xml
      <add>
        <doc>
          <field name="type">source</field>
          <field name="name">kcmodule.cpp</field>
          <field name="url">kcmodule_8cpp_source.html#l00001</field>
          <field name="keywords"></field>
          <field name="text"></field>
        </doc>
      </add>
    """

    doclist = []
    tree = xmlET.parse(lib.outputdir + '/searchdata.xml')
    for doc_child in tree.getroot():
        field = {}
        for child in doc_child:
            if child.attrib['name'] == "type":
                if child.text == 'source':
                    field = None
                    break  # We go to next <doc>
                field['type'] = child.text
            elif child.attrib['name'] == "name":
                field['name'] = child.text
            elif child.attrib['name'] == "url":
                field['url'] = child.text
            elif child.attrib['name'] == "keywords":
                field['keyword'] = child.text
            elif child.attrib['name'] == "text":
                field['text'] = "" if child.text is None else child.text
        if field is not None:
            doclist.append(field)

    indexdic = {
        'name': lib.name,
        'fancyname': lib.fancyname,
        'docfields': doclist
        }

    with open(lib.outputdir + '/html/searchdata.json', 'w') as f:
        for chunk in json.JSONEncoder().iterencode(indexdic):
            f.write(chunk)


def create_product_index(product):
    doclist = []
    for lib in product.libraries:
        with open(lib.outputdir+'/html/searchdata.json', 'r') as f:
            libindex = json.load(f)
            for item in libindex['docfields']:
                if lib.part_of_group:
                    item['url'] = lib.name.lower() + '/html/' + item['url']
                else:
                    item['url'] = 'html/' + item['url']
            doclist.append(libindex)

    indexdic = {
        'name': product.name,
        'fancyname': product.fancyname,
        'libraries': doclist
        }

    with open(product.outputdir + '/searchdata.json', 'w') as f:
        for chunk in json.JSONEncoder().iterencode(indexdic):
            f.write(chunk)


def create_global_index(products):
    doclist = []
    for product in products:
        if product.metainfo['qdoc']:
            continue

        with open(product.outputdir+'/searchdata.json', 'r') as f:
            prodindex = json.load(f)
            for proditem in prodindex['libraries']:
                for item in proditem['docfields']:
                    item['url'] = os.path.join(product.name, item['url'])
            doclist.append(prodindex)

    indexdic = {
        'all': doclist
        }
    with open('searchdata.json', 'w') as f:
        for chunk in json.JSONEncoder().iterencode(indexdic):
            f.write(chunk)


def create_qch(products, tagfiles):
    tag_root = "QtHelpProject"
    tag_files = "files"
    tag_filter_section = "filterSection"
    tag_keywords = "keywords"
    tag_toc = "toc"
    for product in products:
        tree_out = ET.ElementTree(ET.Element(tag_root))
        root_out = tree_out.getroot()
        root_out.set("version", "1.0")
        namespace = ET.SubElement(root_out, "namespace")
        namespace.text = "org.kde." + product.name
        virtual_folder = ET.SubElement(root_out, "virtualFolder")
        virtual_folder.text = product.name
        filter_section = ET.SubElement(root_out, tag_filter_section)
        filter_attribute = ET.SubElement(filter_section, "filterAttribute")
        filter_attribute.text = "doxygen"
        toc = ET.SubElement(filter_section, "toc")
        keywords = ET.SubElement(filter_section, tag_keywords)
        if len(product.libraries) > 0:
            if product.libraries[0].part_of_group:
                product_index_section = ET.SubElement(toc, "section", {'ref': product.name + "/index.html", 'title': product.fancyname})
        files = ET.SubElement(filter_section, tag_files)

        for lib in sorted(product.libraries, key=lambda lib: lib.name):
            tree = ET.parse(lib.outputdir + '/html/index.qhp')
            root = tree.getroot()
            for child in root.findall(".//*[@ref]"):
                if lib.part_of_group:
                    child.attrib['ref'] = lib.name + "/html/" + child.attrib['ref']
                else:
                    child.attrib['ref'] = "html/" + child.attrib['ref']
                child.attrib['ref'] = product.name + '/' + child.attrib['ref']

            for child in root.find(".//"+tag_toc):
                if lib.part_of_group:
                    product_index_section.append(child)
                else:
                    toc.append(child)

            for child in root.find(".//keywords"):
                keywords.append(child)

            resources = [
                "*.json",
                product.name + "/*.json",
                "resources/css/*.css",
                "resources/3rd-party/bootstrap/css/*.css",
                "resources/3rd-party/jquery/jquery-3.1.0.min.js",
                "resources/*.svg",
                "resources/js/*.js",
                "resources/icons/*",
            ]
            if product.part_of_group:
                resources.extend([
                    product.name + "/*.html",
                    product.name + "/" + lib.name + "/html/*.html",
                    product.name + "/" + lib.name + "/html/*.png",
                    product.name + "/" + lib.name + "/html/*.css",
                    product.name + "/" + lib.name + "/html/*.js",
                    product.name + "/" + lib.name + "/html/*.json"
                    ])

            else:
                resources.extend([
                    product.name + "/html/*.html",
                    product.name + "/html/*.png",
                    product.name + "/html/*.css",
                    product.name + "/html/*.js"
                    ])

            for resource in resources:
                file_elem = ET.SubElement(files, "file")
                file_elem.text = resource

        if not os.path.isdir('qch'):
            os.mkdir('qch')

        name = product.name+".qhp"
        outname = product.name+".qch"
        tree_out.write(name, encoding="utf-8", xml_declaration=True)

        # On many distributions, qhelpgenerator from Qt5 is suffixed with
        # "-qt5". Look for it first, and fall back to unsuffixed one if
        # not found.
        qhelpgenerator = find_executable("qhelpgenerator-qt5")

        if qhelpgenerator is None:
            qhelpgenerator = "qhelpgenerator"

        subprocess.call([qhelpgenerator, name, '-o', 'qch/'+outname])
        os.remove(name)



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

from contextlib import contextmanager

class Block(object):
    INDENT_SIZE = 4

    def __init__(self, out, depth = 0):
        self.out = out
        self.depth = depth

    def writeln(self, text):
        self.out.write(self.depth * Block.INDENT_SIZE * " " + text + "\n")

    def write_attrs(self, **attrs):
        for key, value in attrs.items():
            self.writeln('"{}" = "{}";'.format(key, value))

    def write_list_attrs(self, name, **attrs):
        with self.square_block(name) as b:
            for key, value in attrs.items():
                b.writeln('"{}" = "{}"'.format(key, value))

    def write_nodes(self, nodes):
        for node in sorted(nodes):
            self.writeln('"{}";'.format(node))

    @contextmanager
    def block(self, opener, closer, **attrs):
        self.writeln(opener)
        block = Block(self.out, depth=self.depth + 1)
        block.write_attrs(**attrs)
        yield block
        self.writeln(closer)

    def square_block(self, prefix, **attrs):
        return self.block(prefix + " [", "]", **attrs)

    def curly_block(self, prefix, **attrs):
        return self.block(prefix + " {", "}", **attrs)

    def cluster_block(self, title, **attrs):
        return self.curly_block("subgraph " + quote("cluster_" + title), label=title, **attrs)


def quote(txt):
    return '"{}"'.format(txt)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Alex Merry <alex.merry@kdemail.net>
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
# SPDX-FileCopyrightText: 2014 Alex Turbov <i.zaufi@gmail.com>
# SPDX-FileCopyrightText: 2016 Olivier Churlaud <olivier@churlaud.com>
#
# SPDX-License-Identifier: BSD-2-Clause

import logging
import codecs
import os
import sys
import time
import datetime

from kapidox import generator, utils, argparserutils, preprocessing, hlfunctions

try:
    from kapidox import depdiagram
    DEPDIAGRAM_AVAILABLE = True
except ImportError:
    DEPDIAGRAM_AVAILABLE = False


def get_identities():
    """Extract account information from the file (if any) specified with
       the `--accountsfile` command-line argument.
       An account file is a whitespace-separated file with a first "column"
       specifying the account name, the last "column" specifying an email
       address, and all the columns in between are a name, e.g.

       adridg Adriaan de Groot groot@kde.org

       This line has three "columns" for the name.
    """
    maintainers = {}

    args = argparserutils.parse_args(DEPDIAGRAM_AVAILABLE)
    if args.accountsfile and os.path.exists(args.accountsfile):
        logging.debug("Found contributor identities file at %s", args.accountsfile)
        with codecs.open(args.accountsfile, 'r', encoding='utf8') as f:
            for line in f:
                parts = line.strip().split()
                if len(parts) >= 3:
                    maintainers[parts[0]] = {
                        'name': ' '.join(parts[1:-1]),
                        'email': parts[-1]
                        }

    else:
        logging.debug("No contributor identities file specified. Without the "
                      "file, maintainer information will be incomplete. For "
                      "generating KDE API documentation you can download the "
                      "KDE accounts file from https://websvn.kde.org/*check"
                      "out*/trunk/kde-common/accounts and specify it with the "
                      "--accountsfile argument.")

    return maintainers


def main():
    kde_copyright = '1996-' + str(datetime.date.today().year) + ' The KDE developers'
    get_maintainers = get_identities

    hlfunctions.do_it(maintainers_fct=get_maintainers,
                      copyright=kde_copyright)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2016 Olivier Churlaud <olivier@churlaud.com>
#
# SPDX-License-Identifier: BSD-2-Clause

import logging
import os.path
import string
from kapidox import utils

## @package kapidox.models
#
# Contains the classes representing the objects used by kapidox
#

class Library(object):
    """ Library
    """

    def __init__(self, metainfo, products, platforms, all_maintainers):
        """
            Constructor of the Library object
            
            Args:
                metainfo:     (dict) dictionary describing a library
                products:     (list of Products) list of all already created products
                platforms:    (dict) dictionary of all platforms for which the library
                              is available, where the key is a platform and the value
                              is a restriction. For instance:  
                                { 
                                    'Linux': '', 
                                    'Windows': 'Tested with Windows 10 only'
                                }
                              would work.
                all_maintainers: (dict of dict)  all possible maintainers, where the main key
                              is a username/unique pseudo, and the key is a dictionary of name,
                              email address. For example:
                                {
                                    'username01': { 'name': 'Paul Developer', 'email': 'mail@example.com' },
                                    'username02': { 'name': 'Marc Developer2', 'email': 'mail2@example.com' }
                                }
                              would work. 
                                
        """
        self.product = None        
        self.subproduct = None

        if 'group' in metainfo:
            productname = metainfo['group']
            self.part_of_group = True
        else:
            productname = metainfo['name']
            self.part_of_group = False
        if utils.serialize_name(productname) not in products:
            productname = metainfo['name']
            del metainfo['group']
            products[utils.serialize_name(metainfo['name'])] = Product(metainfo, all_maintainers)
            self.part_of_group = False
            logging.warning("Group of {fancyname} not found: dropped.".format_map(metainfo))
        self.product = products[utils.serialize_name(productname)]
        if self.product is None:
            raise ValueError("'{name}' does not belong to a product."
                             .format_map(metainfo))

        if 'subgroup' in metainfo and self.part_of_group:
            for sp in self.product.subproducts:
                if sp.name == utils.serialize_name(metainfo['subgroup']):
                    self.subproduct = sp
            if self.subproduct is None:
                logging.warning("Subgroup {subgroup} of library {name} not documented, subgroup will be None"
                                .format_map(metainfo))

        if self.subproduct is not None:
            self.parent = self.subproduct
            self.subproduct.libraries.append(self)
        else:
            self.parent = self.product
        self.product.libraries.append(self)

        self.metainfo = metainfo
        self.name = metainfo['name']
        self.fancyname = metainfo['fancyname']
        self.description = metainfo.get('description')
        self.maintainers = utils.set_maintainers(metainfo.get('maintainer'), all_maintainers)
        self.platforms = platforms
        self.outputdir = self._set_outputdir(self.part_of_group)
        self.href = '../' + self.outputdir.lower() + '/html/index.html'
        self.path = metainfo['path']
        self.srcdirs = utils.tolist(metainfo.get('public_source_dirs', ['src']))
        self.docdir = utils.tolist(metainfo.get('public_doc_dir', ['docs']))
        if 'public_example_dirs' in metainfo:
            self.exampledirs = utils.tolist(metainfo.get('public_example_dirs', ['examples']))
        else:
            # backward compat
            self.exampledirs = utils.tolist(metainfo.get('public_example_dir', ['examples']))
        self.dependency_diagram = None
        self.type = metainfo.get('type', '')
        self.portingAid = metainfo.get('portingAid', False)
        self.deprecated = metainfo.get('deprecated', False)
        self.libraries = metainfo.get('libraries', [])
        self.cmakename = metainfo.get('cmakename', '')
        self.irc = metainfo.get('irc', self.product.irc)
        self.mailinglist = metainfo.get('mailinglist', self.product.mailinglist)
        self.repopath = utils.set_repopath(metainfo['repo_id'])

    def _extend_parent(self, metainfo, key, key_obj, default):
        if key in metainfo:
            return metainfo[key]
        elif getattr(self.product, key_obj) is not None:
            return getattr(self.product, key_obj)
        else:
            return default

    def _set_outputdir(self, grouped):
        outputdir = self.name
        if grouped:
            outputdir = os.path.join(self.product.outputdir, outputdir)
        return outputdir.lower()


class Product(object):
    """ Product
    """

    # TODO: If no name and no group, it will fail !
    def __init__(self, metainfo, all_maintainers):
        """
            Constructor of the Product object
            
            Args:
                metainfo:     (dict) dictionary describing a product
                all_maintainers: (dict of dict)  all possible maintainers, where the main key
                              is a username/unique pseudo, and the key is a dictionary of name,
                              email address. For example:
                                {
                                    'username01': { 'name': 'Paul Developer', 'email': 'mail@example.com' },
                                    'username02': { 'name': 'Marc Developer2', 'email': 'mail2@example.com' }
                                }
                              would work. 
                                
        """
        
        self.metainfo = metainfo
        self.parent = None
        # if there is a group, the product is the group
        # else the product is directly the library
            
        if 'group_info' in metainfo:
            self.name = utils.serialize_name(metainfo['group_info'].get('name', metainfo.get('group')))
            self.fancyname = metainfo['group_info'].get('fancyname', string.capwords(self.name))
            self.description = metainfo['group_info'].get('description')
            self.long_description = metainfo['group_info'].get('long_description', [])
            self.maintainers = utils.set_maintainers(metainfo['group_info'].get('maintainer'),
                                                     all_maintainers)
            self.platforms = metainfo['group_info'].get('platforms')
            self.outputdir = self.name
            self.href = self.outputdir + '/index.html'
            self.logo_url_src = self._set_logo_src(metainfo['path'],
                                                   metainfo['group_info'])
            self.logo_url = self._set_logo()
            self.libraries = []  # We'll set this later
            self.subgroups = []  # We'll set this later
            self.irc = metainfo['group_info'].get('irc', 'kde-devel')
            self.mailinglist = metainfo['group_info'].get('mailinglist', 'kde-devel')
            self.subproducts = self._extract_subproducts(metainfo['group_info'])
            self.part_of_group = True

        elif 'group' not in metainfo:
            self.name = utils.serialize_name(metainfo['name'])
            self.fancyname = metainfo['fancyname']
            self.description = metainfo.get('description')
            self.maintainers = utils.set_maintainers(metainfo.get('maintainer'), all_maintainers)
            self.platforms = [x['name'] for x in metainfo.get('platforms', [{'name': None}])]
            self.outputdir = self.name
            self.href = self.outputdir + '/html/index.html'
            self.logo_url_src = self._set_logo_src(metainfo['path'], metainfo)
            self.logo_url = self._set_logo()
            self.libraries = []
            self.irc = None
            self.mailinglist = None
            self.part_of_group = False
        else:
            raise ValueError("I do not recognize a product in {name}."
                             .format_map(metainfo))

    def _extract_subproducts(self, groupinfo):
        subproducts = []
        if 'subgroups' in groupinfo:
            for sg in groupinfo['subgroups']:
                if 'name' in sg:
                    subproducts.append(Subproduct(sg, self))
        return subproducts

    def _set_logo(self):
        if self.logo_url_src is not None:
            filename, ext = os.path.splitext(self.logo_url_src)
            return os.path.join(self.outputdir, self.name) + ext
        else:
            return None

    def _set_logo_src(self, path, dct):
        defined_not_found = False
        if 'logo' in dct:
            logo_url = os.path.join(path, dct['logo'])
            if os.path.isfile(logo_url):
                return logo_url
            else:
                defined_not_found = True

        logo_url = os.path.join(path, 'logo.png')
        if os.path.isfile(logo_url):
            if defined_not_found:
                logging.warning(f'Defined {self.fancyname} logo file does not exist, set back to found logo.png')
            return logo_url

        if defined_not_found:
            logging.warning(f'Defined {self.fancyname} logo file does not exist, set back to None')

        return None


class Subproduct(object):
    """ Subproduct
    """
    def __init__(self, spinfo, product):
        """
            Constructor of the Subproduct object
            
            Args:
                spinfo:       (dict) description of the subproduct. It is not more than:
                {
                    'name': 'Subproduct Name',
                    'description': 'This subproduct does this and that',
                    'order': 3, # this is optional
                }
                for example.
                product:      (Product) the product it is part of.
        """
        self.fancyname = spinfo['name']
        self.name = utils.serialize_name(spinfo['name'])
        self.description = spinfo.get('description')
        self.order = spinfo.get('order', 99)  # If no order, go to end
        self.libraries = []
        self.product = product
        self.parent = product



from setuptools import setup, find_packages

setup(
        name='kapidox',
        version='5.116.0',
        description='KDE API documentation generation tools',
        maintainer='Olivier Churlaud',
        maintainer_email='olivier@churlaud.com',
        url='https://invent.kde.org/frameworks/kapidox',
        packages=find_packages(),
        python_requires=">=3.8",
        scripts=['kapidox/depdiagram_generate_all'],
        entry_points={
            "console_scripts": [
                "kapidox-generate = kapidox.kapidox_generate:main",
                "kapidox-depdiagram-prepare = kapidox.depdiagram_prepare:main",
                "kapidox-depdiagram-generate = kapidox.depdiagram_generate:main",
            ],
        },
        install_requires=["doxypypy", "doxyqml", "requests", "jinja2", "pyyaml"]
    )



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2016 Olivier Churlaud <olivier@churlaud.com>
# SPDX-FileCopyrightText: 2014 Alex Merry <alex.merry@kdemail.net>
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
# SPDX-FileCopyrightText: 2014 Alex Turbov <i.zaufi@gmail.com>
#
# SPDX-License-Identifier: BSD-2-Clause

import logging
import os
import sys
from typing import Any, Dict, Optional

from urllib.request import Request, urlopen
from urllib.error import HTTPError

import yaml

from kapidox import utils
from kapidox.models import Library, Product

__all__ = (
    "create_metainfo",
    "parse_tree")

PLATFORM_ALL = "All"
PLATFORM_UNKNOWN = "UNKNOWN"


## @package kapidox.preprocessing
#
# Preprocessing of the needed information.
#
# The module allow to walk through folders, read metainfo files and create
# products, subgroups and libraries representing the projects.
#

def expand_platform_all(dct, available_platforms):
    """If one of the keys of dct is `PLATFORM_ALL` (or `PLATFORM_UNKNOWN`),
    remove it and add entries for all available platforms to dct

    Args:
        dct: (dictionary) dictionary to expand
        available_platforms: (list of string) name of platforms
    """

    add_all_platforms = False
    if PLATFORM_ALL in dct:
        note = dct[PLATFORM_ALL]
        add_all_platforms = True
        del dct[PLATFORM_ALL]
    if PLATFORM_UNKNOWN in dct:
        add_all_platforms = True
        note = dct[PLATFORM_UNKNOWN]
        del dct[PLATFORM_UNKNOWN]
    if add_all_platforms:
        for platform in available_platforms:
            if platform not in dct:
                dct[platform] = note


def create_metainfo(path) -> Optional[Dict[str, Any]]:
    """Look for a `metadata.yaml` file and create a dictionary out it.

    Args:
        path: (string) the current path to search.
    Returns:
        A dictionary containing all the parsed information, or `None` if it
    did not fulfill some conditions.
    """

    metainfo: Optional[Dict[str, Any]]

    if not os.path.isdir(path):
        return None

    try:
        metainfo_file = os.path.join(path, 'metainfo.yaml')
    except UnicodeDecodeError as e:
        logging.warning('Unusual base path {!r} for metainfo.yaml'.format(path))
        return None
    if not os.path.isfile(metainfo_file):
        return None

    try:
        metainfo = yaml.safe_load(open(metainfo_file))
    except Exception as e:
        print(e)
        logging.warning(f'Could not load metainfo.yaml for {path}, skipping it')
        return None

    if metainfo is None:
        logging.warning(f'Empty metainfo.yaml for {path}, skipping it')
        return None

    if 'subgroup' in metainfo and 'group' not in metainfo:
        logging.warning(f'Subgroup but no group in {path}, skipping it')
        return None

    # Suppose we get a relative path passed in (e.g. on the command-line,
    # path .. because we're building the dox in a subdirectory of a source
    # checkout) then we don't want dirname to be "..", but the name that
    # that resolves to.
    dirname = os.path.basename(os.path.abspath(path))
    if 'fancyname' in metainfo:
        fancyname = metainfo['fancyname']
    else:
        fancyname = utils.parse_fancyname(path)

    if not fancyname:
        logging.warning(f'Could not find fancy name for {path}, skipping it')
        return None
    # A fancyname has 1st char capitalized
    fancyname = fancyname[0].capitalize() + fancyname[1:]

    if 'repo_id' in metainfo:
        repo_id = metainfo['repo_id']
    else:
        repo_id = dirname

    qdoc: bool = False

    if 'qdoc' in metainfo:
        qdoc = metainfo['qdoc']

    metainfo.update({
        'fancyname': fancyname,
        'name': dirname,
        'repo_id': repo_id,
        'public_lib': metainfo.get('public_lib', False),
        'dependency_diagram': None,
        'path': path,
        'qdoc': qdoc,
    })

    # replace legacy platform names
    if 'platforms' in metainfo:
        platforms = metainfo['platforms']
        for index, x in enumerate(platforms):
            if x['name'] == "MacOSX":
                x['name'] = "macOS"
                platforms[index] = x
                logging.warning('{fancyname} uses outdated platform name, please replace "MacOSX" with "macOS".'
                                .format_map(metainfo))
        metainfo.update({'platforms': platforms})
    if 'group_info' in metainfo:
        group_info = metainfo['group_info']
        if 'platforms' in group_info:
            platforms = group_info['platforms']
            for index, x in enumerate(platforms):
                if "MacOSX" in x:
                    x = x.replace("MacOSX", "macOS")
                    platforms[index] = x
                    logging.warning('Group {fancyname} uses outdated platform name, please replace "MacOSX" with "macOS".'
                                    .format_map(group_info))
            group_info.update({'platforms': platforms})

    return metainfo


def parse_tree(rootdir):
    """Recursively call create_metainfo() in subdirs of rootdir

    Args:
        rootdir: (string)  Top level directory containing the libraries.

    Returns:
        A list of metainfo dictionary (see create_metainfo()).

    """
    metalist = []
    for path, dirs, _ in os.walk(rootdir, topdown=True):
        # We don't want to do the recursion in the dotdirs
        dirs[:] = [d for d in dirs if not d[0] == '.']
        metainfo = create_metainfo(path)
        if metainfo is not None:
            # There was a metainfo.yaml, which means it was
            # the top of a checked-out repository. Stop processing,
            # because we do not support having repo B checked out (even
            # as a submodule) inside repo A.
            #
            # There are exceptions: messagelib (KDE PIM) contains
            # multiple subdirectories with their own metainfo.yaml,
            # which are listed as public sources.
            dirs[:] = [d for d in dirs if d in metainfo.get('public_source_dirs', [])]
            if metainfo['public_lib'] or 'group_info' in metainfo:
                metalist.append(metainfo)
            else:
                logging.warning('{name} has no public libraries'.format_map(metainfo))

    return metalist


def sort_metainfo(metalist, all_maintainers):
    """Extract the structure (Product/Subproduct/Library) from the metainfo
    list.

    Args:
        metalist: (list of dict) lists of the metainfo extracted in parse_tree().
        all_maintainers: (dict of dict)  all possible maintainers.

    Returns:
        A list of Products, a list of groups (which are products containing
    several libraries), a list of Libraries and the available platforms.
    """
    products = dict()
    groups = []
    libraries = []
    available_platforms = {'Windows', 'macOS', 'Linux', 'Android', 'FreeBSD'}

    # First extract the structural info
    for metainfo in metalist:
        product = extract_product(metainfo, all_maintainers)
        if product is not None:
            products[product.name] = product

    # Second extract the libraries
    for metainfo in metalist:
        try:
            platforms = metainfo['platforms']
            platform_lst = [x['name'] for x in platforms
                            if x['name'] not in (PLATFORM_ALL,
                                                 PLATFORM_UNKNOWN)]

            available_platforms.update(set(platform_lst))
        except (KeyError, TypeError):
            logging.warning('{fancyname} library lacks valid platform definitions'
                            .format_map(metainfo))
            platforms = [dict(name=PLATFORM_UNKNOWN)]

        dct = dict((x['name'], x.get('note', '')) for x in platforms)

        expand_platform_all(dct, available_platforms)
        platforms = dct

        if metainfo['public_lib']:
            lib = Library(metainfo, products, platforms, all_maintainers)
            libraries.append(lib)

    groups = []
    for key in products.copy():
        if len(products[key].libraries) == 0:
            del products[key]
        elif products[key].part_of_group:
            groups.append(products[key])

    return list(products.values()), groups, libraries, available_platforms


def extract_product(metainfo, all_maintainers):
    """Extract a product from a metainfo dictionary.

    Args:
        metainfo: (dict) metainfo created by the create_metainfo() function.
        all_maintainers: (dict of dict) all possible maintainers

    Returns:
        A Product or None if the metainfo does not describe a product.
    """

    if 'group_info' not in metainfo and 'group' in metainfo:
        # This is not a product but a simple lib
        return None

    try:
        product = Product(metainfo, all_maintainers)
        return product
    except ValueError as e:
        logging.error(e)
        return None



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Alex Merry <alex.merry@kdemail.net>
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
# SPDX-FileCopyrightText: 2014 Alex Turbov <i.zaufi@gmail.com>
# SPDX-FileCopyrightText: 2016 Olivier Churlaud <olivier@churlaud.com>
#
# SPDX-License-Identifier: BSD-2-Clause

import logging
import os
import shutil
import sys
import tempfile
import json

from urllib.request import urlretrieve

from . import generator, utils, argparserutils, preprocessing

try:
    from kapidox import depdiagram

    DEPDIAGRAM_AVAILABLE = True
except ImportError:
    DEPDIAGRAM_AVAILABLE = False


def do_it(maintainers_fct, copyright, searchpaths=None):
    utils.setup_logging()
    if searchpaths is None:
        searchpaths = searchpaths = ['/usr/share/doc/qt5', '/usr/share/doc/qt']
    args = argparserutils.parse_args(DEPDIAGRAM_AVAILABLE)

    if len(os.listdir(os.getcwd())) > 0:
        logging.error("Run this command from an empty directory.")
        exit(2)

    if not DEPDIAGRAM_AVAILABLE:
        logging.warning("Missing Graphviz dependency: diagrams will not be generated.")

    tagfiles = generator.search_for_tagfiles(
        suggestion=args.qtdoc_dir,
        doclink=args.qtdoc_link,
        flattenlinks=args.qtdoc_flatten_links,
        searchpaths=searchpaths)

    rootdir = args.sourcesdir
    maintainers = maintainers_fct()

    metalist = preprocessing.parse_tree(rootdir)
    products, groups, libraries, available_platforms = preprocessing.sort_metainfo(metalist, maintainers)

    dirsrc = os.path.join(args.doxdatadir, 'htmlresource')
    dirdest = 'resources'
    if os.path.isdir(dirdest):
        shutil.rmtree(dirdest)
    shutil.copytree(dirsrc, dirdest)
    os.rename(dirdest + '/favicon.ico', './favicon.ico')
    os.rename(dirdest + '/worker.js', './worker.js')

    generator.process_toplevel_html_file('index.html',
                                         args.doxdatadir,
                                         title=args.title,
                                         products=products,
                                         qch_enabled=args.qhp
                                         )
    generator.process_subgroup_html_files('index.html',
                                          args.doxdatadir,
                                          title=args.title,
                                          groups=groups,
                                          available_platforms=available_platforms,
                                          qch_enabled=args.qhp
                                          )
    tmp_dir = tempfile.mkdtemp(prefix='kapidox-')

    try:
        if args.depdiagram_dot_dir:
            dot_files = utils.find_dot_files(args.depdiagram_dot_dir)
            assert dot_files
        for lib in libraries:
            logging.info(f'# Generating doc for {lib.fancyname}')
            if args.depdiagram_dot_dir:
                png_path = os.path.join(tmp_dir, lib.name) + '.png'
                ok = generator.generate_diagram(png_path, lib.fancyname,
                                                dot_files, tmp_dir)
                if ok:
                    lib.dependency_diagram = png_path

            # store this as we won't use that every time
            create_qhp = args.qhp
            args.qhp = False
            ctx = generator.create_fw_context(args, lib, tagfiles)
            # set it back
            args.qhp = create_qhp

            generator.gen_fw_apidocs(ctx, tmp_dir)
            tagfiles.insert(0, generator.create_fw_tagfile_tuple(lib))

        # Rebuild for interdependencies
        for lib in libraries:
            logging.info(f'# Rebuilding {lib.fancyname} for interdependencies')
            shutil.rmtree(lib.outputdir)
            ctx = generator.create_fw_context(args, lib, tagfiles, copyright)
            generator.gen_fw_apidocs(ctx, tmp_dir)
            generator.finish_fw_apidocs(ctx)
            if not ctx.is_qdoc:
                logging.info('# Generate indexing files')
                generator.indexer(lib)
        for product in products:
            if not product.metainfo['qdoc']:
                generator.create_product_index(product)

            if product.logo_url is not None:
                logodir = os.path.dirname(product.logo_url)
                if not os.path.isdir(logodir):
                    os.mkdir(logodir)
                shutil.copy(product.logo_url_src, product.logo_url)
        generator.create_global_index(products)
        if args.qhp:
            logging.info('# Merge qch files')
            generator.create_qch(products, tagfiles)
        logging.info("# Writing metadata...")
        with open('metadata.json', 'w') as file:
            json.dump(metalist, file)
        libs = []
        for lib in libraries:
            libs.append({"name": lib.name, "outputdir": lib.outputdir})
        with open('outputs.json', 'w') as file:
            json.dump(libs, file)
        logging.info('# Done')
    finally:
        if args.keep_temp_dirs:
            logging.info(f'Kept temp dir at {tmp_dir}')
        else:
            shutil.rmtree(tmp_dir)



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2014 Aurélien Gâteau <agateau@kde.org>
#
# SPDX-License-Identifier: BSD-2-Clause

import logging
import itertools
from functools import cmp_to_key

from kapidox.depdiagram.block import Block, quote
from kapidox.depdiagram.framework import Framework
from kapidox.depdiagram.frameworkdb import FrameworkDb

__all__ = ('generate',)

ROOT_NODE_ATTRS = dict(fontsize=12, shape="box")

GROUP_ATTRS = dict(style="filled", fillcolor="grey95", color="grey85")

# Blocks within groups
OTHER_ATTRS = dict(style="filled", fillcolor="cornsilk", color="black")

QT_ATTRS = dict(style="filled", fillcolor="darkseagreen1", color="black")

FW_ATTRS = dict(style="filled", fillcolor="azure", color="black")

# Target blocks, used with --detailed
FW_TARGET_ATTRS = dict(style="filled", fillcolor="paleturquoise")

# Highlight the wanted framework, used with --framework
WANTED_FW_ATTRS = dict(penwidth=2)

class FrameworkCmp(object):
    def __init__(self, db):
        self.db = db
        self.src = set(db)
        self.dst = []

    def __call__(self, fw1, fw2):
        if self.depends_on(fw1, fw2):
            return 1
        if self.depends_on(fw2, fw1):
            return -1
        return 0

    def depends_on(self, depender_fw, provider_fw):
        for dep_target in depender_fw.get_all_target_dependencies():
            if provider_fw.has_target(dep_target):
                return True

            try:
                dep_fw = self.db.get_framework_for_target(dep_target)
            except KeyError:
                # No framework for this target, must be an external dependency,
                # carry on
                continue
            if dep_fw != depender_fw and self.depends_on(dep_fw, provider_fw):
                return True

        return False


class DotWriter(Block):
    def __init__(self, db, out, wanted_fw=None, detailed=False):
        Block.__init__(self, out)
        self.db = db
        self.detailed = detailed
        self.wanted_fw = wanted_fw

    def write(self):
        with self.curly_block("digraph Root") as root:
            root.write_list_attrs("node", **ROOT_NODE_ATTRS)

            other_targets = self.db.find_external_targets()
            qt_targets = set([x for x in other_targets if x.startswith("Qt")])
            other_targets.difference_update(qt_targets)

            if qt_targets:
                with root.cluster_block("Qt", **GROUP_ATTRS) as b:
                    b.write_list_attrs("node", **QT_ATTRS)
                    b.write_nodes(qt_targets)

            if other_targets:
                with root.cluster_block("Others", **GROUP_ATTRS) as b:
                    b.write_list_attrs("node", **OTHER_ATTRS)
                    b.write_nodes(other_targets)

            lst = sorted([x for x in self.db], key=lambda x: x.tier)
            for tier, frameworks in itertools.groupby(lst, lambda x: x.tier):
                cluster_title = "Tier {}".format(tier)
                with root.cluster_block(cluster_title, **GROUP_ATTRS) as tier_block:
                    tier_block.write_list_attrs("node", **FW_ATTRS)
                    # Sort frameworks within the tier to ensure frameworks which
                    # depend on other frameworks from that tier are listed after
                    # their dependees.
                    frameworks = list(frameworks)
                    for fw in sorted(frameworks, key=cmp_to_key(FrameworkCmp(self.db))):
                        if self.detailed:
                            self.write_detailed_framework(tier_block, fw)
                        else:
                            self.write_framework(tier_block, fw)

    def write_framework(self, tier_block, fw):
        if fw == self.wanted_fw:
            tier_block.write_list_attrs(quote(fw.name), **WANTED_FW_ATTRS)
        else:
            tier_block.write_nodes([fw.name])
        edges = set([])
        for target in fw.get_all_target_dependencies():
            try:
                target_fw = self.db.get_framework_for_target(target)
                if fw == target_fw:
                    continue
                dep = target_fw.name
            except KeyError:
                dep = target
            edges.add((fw.name, dep))
        for dep_fw in fw.get_extra_frameworks():
            edges.add((fw.name, dep_fw))
        for edge in edges:
            tier_block.writeln('"{}" -> "{}";'.format(*edge))

    def write_detailed_framework(self, tier_block, fw):
        with tier_block.cluster_block(fw.name, **FW_ATTRS) as fw_block:
            if fw == self.wanted_fw:
                fw_block.write_attrs(**WANTED_FW_ATTRS)
            fw_block.write_list_attrs("node", **FW_TARGET_ATTRS)
            targets = sorted(fw.get_targets())
            fw_block.write_nodes(targets)
            for target in targets:
                deps = fw.get_dependencies_for_target(target)
                for dep in sorted(deps):
                    fw_block.writeln('"{}" -> "{}";'.format(target, dep))


def generate(out, dot_files, framework=None, with_qt=False, detailed=False):
    db = FrameworkDb()
    db.populate(dot_files, with_qt=with_qt)

    if framework:
        wanted_fw = db.find_by_name(framework)
        if wanted_fw is None:
            logging.error("No framework named {}.".format(framework))
            return False
        db.remove_unused_frameworks(wanted_fw)
    else:
        wanted_fw = None

    writer = DotWriter(db, out, wanted_fw=wanted_fw, detailed=detailed)
    writer.write()

    return True

# vi: ts=4 sw=4 et
